#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>

#include "ckc.h"

struct srcfile {
	txt			name;
	txt			body;
	struct stat		stat;
	int			fd;
	TAILQ_ENTRY(srcfile)	list;
	struct chunk		chunk;
};


static TAILQ_HEAD(, srcfile)	srcfiles = TAILQ_HEAD_INITIALIZER(srcfiles);

/*********************************************************************
 * chunkmethods
 */

static void
srcfile_quote(FILE *f, struct chunk *c, txt t)
{
	struct srcfile *sf;
	const char *p, *q, *r;

	while (t.b < t.e && ct_white(*t.b))
		t.b++;
	while (t.b < t.e && ct_white(t.e[-1]))
		t.e--;
	sf = c->priv;
	fprintf(f, "file \"%.*s\"\n", PT(sf->name));
	for (p = t.b; p > sf->body.b && *p != '\n'; p--)
		continue;
	if (*p == '\n')
		p++;
	for (q = t.e; q < sf->body.e && *q != '\n'; q++)
		continue;
	fprintf(f, "\t%.*s\n\t", q - p, p);
	for (r = p; r < t.e; r++) {
		if (ct_white(*r))
			fprintf(f, "%c", *r);
		else if (r < t.b)
			fprintf(f, " ");
		else 
			fprintf(f, "^");
	}
	printf("\n");
	
}

static struct chunk_method srcfile_methods = {
	.name = "srcfile",
	.quote = srcfile_quote
};

/*********************************************************************
 * Load from a fd
 * Return ptr or NULL+errno
 */

struct chunk *
LoadSrcFd(const char *name, int fd)
{
	struct srcfile	*sf, *sf2;
	void *p;
	char *dst;
	int rd;
	size_t sofar, alloced;

	sf = calloc(sizeof *sf, 1);
	AAP(sf);
	sf->chunk.priv = sf;
	sf->chunk.t = &sf->body;
	sf->chunk.f = &srcfile_methods;
	sf->name = Txt(name, NULL);
	sf->fd = fd;
	assert(fstat(fd, &sf->stat) == 0);
	TAILQ_FOREACH(sf2, &srcfiles, list) {
		/* Check if we have the file already */
		if (sf2->stat.st_ino != sf->stat.st_ino)
			continue;
		if (sf2->stat.st_dev != sf->stat.st_dev)
			continue;

		/* We have the file, but has it changed ? */
		if (sf2->stat.st_mtime != sf->stat.st_mtime)
			continue;

		/* Also check size, for good measure */
		if (sf2->stat.st_size != sf->stat.st_size)
			continue;

		/* If we used the same name, use the same instance */
		if (TxtEq(sf->name, sf2->name)) {
			/* XXX: close the fd ? */
			free(sf);
			return (&sf2->chunk);
		} 

		/* Else just reuse the content */
		/* XXX: close the fd ? */
		sf->body = sf2->body;
		TAILQ_INSERT_TAIL(&srcfiles, sf, list);
		return (&sf->chunk);
	}

	if (S_ISREG(sf->stat.st_mode)) {
		/* Try mmap'ing regular files */
		sofar = (size_t)sf->stat.st_size;
		assert((off_t)sofar == sf->stat.st_size);
		p = mmap(NULL, sofar, PROT_READ, 0, fd, 0);
		if (p != MAP_FAILED) {
			sf->body.b = p;
			sf->body.e = sf->body.b + sofar;
			TAILQ_INSERT_TAIL(&srcfiles, sf, list);
			return (&sf->chunk);
		}
	}
	alloced = 256 * 1024;
	dst = malloc(alloced);
	AAP(dst);
	for (sofar = 0; ; ) {
		rd = read(fd, dst + sofar, alloced - sofar);
		assert(rd >= 0);
		if (rd == 0)
			break;
		sofar += (unsigned)rd;
		if (sofar == alloced) {
			alloced += alloced;
			dst = realloc(dst, alloced);
			AAP(dst);
		}
	}
	dst = realloc(dst, sofar);
	AAP(dst);
	sf->body = Txt(dst, dst + sofar);
	TAILQ_INSERT_TAIL(&srcfiles, sf, list);
	return (&sf->chunk);
}

/*********************************************************************
 * Load a file
 * Return ptr or NULL+errno
 */

struct chunk *
LoadSrcFile(const char *pfx, const char *name)
{
	int fd, e;
	char *n;

	if (!strcmp(name, "-"))
		return (LoadSrcFd("<stdin>", 0));

	if (pfx == NULL)
		n = strdup(name);
	else
		asprintf(&n, "%s/%s", pfx, name);

	fd = open(n, O_RDONLY);
	if (fd < 0) {
		e = errno;
		warn("Cannot open \"%s\"", n);
		free (n);
		errno = e;
		return (NULL);
	}
	return (LoadSrcFd(n, fd));
}

/*********************************************************************
 * Load a file
 * Return ptr or NULL+errno
 */

struct chunk *
LoadSrcFileTxt(const char *pfx, txt name)
{
	int fd, e;
	char *n;

	if (pfx == NULL)
		asprintf(&n, "%.*s", PT(name));
	else
		asprintf(&n, "%s/%.*s", pfx, PT(name));

	fd = open(n, O_RDONLY);
	if (fd < 0) {
		free (n);
		return (NULL);
	}
	return (LoadSrcFd(n, fd));
}
