/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: configsource.c,v 1.19 2007/08/22 09:06:22 phk Exp $
 *
 * Parse the configuration and get things going.
 *
 */

#include <stdio.h>
#include <sys/queue.h>
#include <stdlib.h>
#include <assert.h>
#include <ctype.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <string.h>

#include <phk_argv.h>

#include "configkit.h"
#include "configkit_src.h"
#include "configkit_int.h"

/*
 * Config sources
 */

struct ck_src;

static LIST_HEAD(, ck_src) ck_src_head =
	LIST_HEAD_INITIALIZER(ck_src_head);


int
ck_printf(CK_ARG, const char *fmt, ...)
{
	va_list ap;
	int i;

	if (cfa == NULL)
		return (0);
	cfa->output_happened++;
	va_start(ap, fmt);
	i = cfa->cfs->print(CK_PASS, fmt, ap);
	va_end(ap);
	return (i);
}

struct ck_src *
ck_SrcCreate(struct ck_tree *cft, ck_printf_t *func, void *priv)
{
	struct ck_src *cf;

	cf = calloc(sizeof *cf, 1);
	LIST_INSERT_HEAD(&ck_src_head, cf, list);
	cf->cft = cft;
	cf->print = func;
	cf->srcpriv = priv;
	cf->cfa.cfs = cf;
	cf->cfa.src = cf->srcpriv;
	return (cf);
}

void
ck_SrcDestroy(struct ck_src *cf)
{

	ck_SrcFeed(cf, "\n", 1);
	LIST_REMOVE(cf, list);
	if (cf->buf != NULL)
		free(cf->buf);
	free(cf);
}

static void
Configure(struct ck_src *cf, char *line)
{
	int i, flag;
	char **av;

	cf->cfa.line = strdup(line);
	av = PhkArgvParse(line, 1);
	assert(av != NULL);
	if (av[0] != NULL) {
		ck_printf(&cf->cfa, "Error: %s\n", av[0]);
		free(cf->cfa.line);
		cf->cfa.line = NULL;
		PhkArgvFree(av);
		return;
	}
	if (av[1] == NULL) {
		free(cf->cfa.line);
		cf->cfa.line = NULL;
		PhkArgvFree(av);
		return;
	}

	cf->tok = av + 1;
	for (cf->ntok = 0; cf->tok[cf->ntok] != NULL; cf->ntok++)
		continue;

	flag = 0;
	/* If the last token is a question mark, treat this as a help request */
	if (!strcmp(cf->tok[cf->ntok - 1], "?")) {
		cf->tok[cf->ntok - 1] = NULL;
		flag |= CK_HELP;
	}

	cf->cfa.cft = cf->cft;
	cf->cfa.tok = cf->tok;
	cf->cfa.flag = flag;
	cf->cfa.tree = strdup("");
	i = __ck_call(&cf->cfa);
	if (i > 0)
		ck_printf(&cf->cfa, "Error: %s\n", strerror(i));
	free(cf->cfa.tree);
	free(cf->cfa.line);
	PhkArgvFree(av);
	return;
}

/*
 * XXX: line editing, command history etc. belongs here.
 * XXX: telnet handlling on the other hand does not(!)
 */
void
ck_SrcFeed(struct ck_src *cf, const void *buf, int len)
{
	char *p, *s, *q;
	int l;

	if (cf->buf != NULL)
		l = strlen(cf->buf);
	else
		l = 0;
	p = realloc(cf->buf, l + len + 1);
	cf->buf = NULL;
	if (p == NULL) {
		return;
	}
	memcpy(p + l, buf, len);
	p[l + len] = '\0';
	q = p;
	for(;;) {
		s = strchr(q, '\n');
		if (s == NULL)
			s = strchr(q, '\r');
		if (s == NULL)
			break;
		*s++ = '\0';
		Configure(cf, q);
		ck_printf(&cf->cfa, NULL);
		q = s;
	}
	if (*q != '\0') {
		/* leave leftovers */
		cf->buf = strdup(q);
	}
	free(p);
}

static int
ck_FILE_printf(CK_ARG, const char *fmt, va_list ap)
{
	FILE *f;

	f = cfa->src;
	assert(f != NULL);
	if (fmt == NULL) {
		fflush(f);
		return (0);
	} else {
		return (vfprintf(f, fmt, ap));
	}
}

static int
ck_FD_printf(CK_ARG, const char *fmt, va_list ap)
{
	int *fdp;
	char *p;
	int i;

	fdp = cfa->src;
	if (fmt == NULL)
		return (0);
	vasprintf(&p, fmt, ap);
	i = write(*fdp, p, strlen(p));
	free(p);
	return(i);
}

void
ck_SrcFd(struct ck_tree *cft, FILE *fout, int fd)
{
	char buf[BUFSIZ];
	int i;
	struct ck_src *cf;

	if (fout != NULL)
		cf = ck_SrcCreate(cft, ck_FILE_printf, fout);
	else
		cf = ck_SrcCreate(cft, ck_FD_printf, &fd);
	while (1) {
		i = read(fd, buf, sizeof buf);
		if (i > 0)
			ck_SrcFeed(cf, buf, i);
		else
			break;
	}
	ck_SrcDestroy(cf);
	return;
}
