/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: convenience.c,v 1.17 2008/04/30 05:58:33 phk Exp $
 *
 * Convenience functions for users
 *
 */
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "configkit.h"

#include "phk_expr.h"

static void
quote(CK_ARG, const char *str)
{
	const char *p;

	for (p = str; *p; p++) {
		if (isspace(*p) || *p == '\\' || *p == '"')
			break;
	}
	if (!*p) {
		ck_printf(CK_PASS, "%s", str);
		return;
	}
	ck_printf(CK_PASS, "\"");
	for (p = str; *p; p++) {
		if (*p == '\\')
			ck_printf(CK_PASS, "\\\\");
		else if (*p == '\n')
			ck_printf(CK_PASS, "\\n");
		else if (*p == '\r')
			ck_printf(CK_PASS, "\\r");
		else if (*p == '\t')
			ck_printf(CK_PASS, "\t");
		else if (*p == ' ')
			ck_printf(CK_PASS, " ");
		else if (*p == '"')
			ck_printf(CK_PASS, "\\\"");
		else if (isspace(*p))
			ck_printf(CK_PASS, "\\%03o", *p);
		else
			ck_printf(CK_PASS, "%c", *p);
	}
	ck_printf(CK_PASS, "\"");
}


int
ck_WordVar(CK_ARG, const char **arg, char **var)
{

	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		if (*var == NULL && (cfa->flag & CK_SHOW))
			ck_printf(CK_PASS, "%s is unset\n", cfa->tree);
		else if (*var == NULL)
			;
		else {
			ck_printf(CK_PASS, "%s ", cfa->tree);
			quote(CK_PASS, *var);
			ck_printf(CK_PASS, "\n");
		}
		return (0);
	}
	if (cfa->flag & CK_NO) {
		if (*var == NULL)
			return (0);
		free(*var);
		*var = NULL;
	} else {
		free(*var);
		*var = strdup(*arg);
	}
	return (1);
}

int
ck_UintVar(CK_ARG, unsigned *arg, unsigned *var, unsigned defval)
{

	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		if (*var == defval && (cfa->flag & CK_SHOW))
			ck_printf(CK_PASS, "%s is unset (%u)\n",
			    cfa->tree, defval);
		else if (*var == defval)
			;
		else
			ck_printf(CK_PASS, "%s %u\n", cfa->tree, *var);
		return (0);
	}
	if (cfa->flag & CK_NO) {
		if (*var == defval)
			return (0);
		*var = defval;
	} else {
		*var = *arg;
	}
	return (1);
}

static int
issame(double a, double b)
{
	if (isnan(a) && isnan(b))
		return (1);
	if (a == b)
		return (1);
	return (0);
}

int
ck_RealVar(CK_ARG, double *arg, double *var, double defval, const char *fmt)
{
	char fmtbuf[BUFSIZ];

	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		if (issame(*var, defval) && (cfa->flag & CK_SHOW)) {
			sprintf(fmtbuf, "%%s is unset (%s)\n", fmt);
			ck_printf(CK_PASS, fmtbuf, cfa->tree, *var);
		} else if (issame(*var, defval)) {
			;
		} else {
			sprintf(fmtbuf, "%%s %s\n", fmt);
			ck_printf(CK_PASS, fmtbuf, cfa->tree, *var);
		}
		return (0);
	}
	if (cfa->flag & CK_NO) {
		if (*var == defval)
			return (0);
		*var = defval;
	} else {
		*var = *arg;
	}
	return (1);
}

int
ck_RealVar2(CK_ARG, const char **arg, double *var, char **strvar, double defval, const char *fmt)
{
	char fmtbuf[BUFSIZ];
	char *ers;

	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		if (*strvar == NULL && (cfa->flag & CK_SHOW)) {
			sprintf(fmtbuf, "%%s is unset (%s)\n", fmt);
			ck_printf(CK_PASS, fmtbuf, cfa->tree, *var);
		} else if (*strvar == NULL) {
			;
		} else {
			sprintf(fmtbuf, "%s %s\t# \"%s\"\n",
			    cfa->tree, fmt, *strvar);
			ck_printf(CK_PASS, fmtbuf, *var);
		}
		return (0);
	}
	free(*strvar);
	*strvar = NULL;
	if (cfa->flag & CK_NO) {
		if (*var == defval)
			return (0);
		*var = defval;
	} else {
		*strvar = strdup(*arg);
		assert(*strvar != NULL);
		*var = phk_expr(*arg, &ers);
		if (ers != NULL)
			ck_printf(CK_PASS, "%s", ers);
	}
	return (1);
}

int
ck_BoolKeyVar(CK_ARG, const char **key, const char *name, unsigned *bits, unsigned bit, unsigned defval)
{

	if (defval)
		defval = bit;

	if (key != NULL && strcmp(*key, name))
		return (0);
	if (cfa->flag & CK_CONF) {
	    	if ((*bits & bit) != defval) 
			ck_printf(CK_PASS, "%s %s\n", cfa->tree, name);
		return (0);
	}
	if (cfa->flag & CK_SHOW) {
	    	if ((*bits & bit) != defval) 
			ck_printf(CK_PASS, "%s %s is enabled\n",
			    cfa->tree, name);
		else
			ck_printf(CK_PASS, "%s %s is disabled\n",
			    cfa->tree, name);
		return (0);
	}
	if (key == NULL)
		return (0);
	if (cfa->flag & CK_NO) {
		*bits &= ~bit;
	} else {
		*bits |= bit;
	}
	return (1);
}

int
ck_BoolVar(CK_ARG, unsigned *bits, unsigned bit, unsigned defval)
{

	if (defval)
		defval = bit;

	if (cfa->flag & CK_CONF) {
	    	if ((*bits & bit) != defval) 
			ck_printf(CK_PASS, "%s\n", cfa->tree);
		return (0);
	}
	if (cfa->flag & CK_SHOW) {
	    	if ((*bits & bit) != defval) 
			ck_printf(CK_PASS, "%s is enabled\n", cfa->tree);
		else
			ck_printf(CK_PASS, "%s is disabled\n", cfa->tree);
		return (0);
	}
	if (cfa->flag & CK_NO) {
		*bits &= ~bit;
	} else {
		*bits |= bit;
	}
	return (1);
}

