/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: libconfigkit.c,v 1.22 2005/12/06 10:20:08 phk Exp $
 *
 * Parse the configuration and get things going.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "configkit.h"
#include "configkit_int.h"

/* Find out if any args before us is the last one */
static int
nulled(CK_ARG, int idx)
{
	int i;

	for (i = 0; i <= idx; i++) {
		if (cfa->tok[i] == NULL &&
		    (cfa->flag & (CK_NO|CK_CONF|CK_SHOW))) {
			return (1);
		}
	}
	return (0);
}

int
__ck_enum(CK_ARG, const char ***dest, int idx, const char **table)
{
	const char **p;

	if (nulled(CK_PASS, idx)) {
		*dest = NULL;
		return (0);
	}
	p = NULL;
	if (cfa->tok[idx] != NULL) {
		for (p = table; *p != NULL; p++) {
			if (strcmp(*p, cfa->tok[idx]))
				continue;
			**dest = *p;
			return (0);
		}
	}
	ck_printf(CK_PASS, "Expected an argument (");
	for (p = table; *p != NULL; p++)
		ck_printf(CK_PASS, " %s", *p);
	ck_printf(CK_PASS, ")\n");
	return (-1);
}

int
__ck_word(CK_ARG, const char ***dest, int idx)
{

	if (nulled(CK_PASS, idx)) {
		*dest = NULL;
		return (0);
	}
	if (cfa->tok[idx] == NULL) {
		ck_printf(CK_PASS,
		    "Expected an argument\n");
		return (-1);
	}
	**dest = cfa->tok[idx];
	return (0);
}

int
__ck_uint(CK_ARG, unsigned **dest, int idx)
{
	unsigned long ul;
	char *ep;

	if (nulled(CK_PASS, idx)) {
		*dest = NULL;
		return (0);
	}
	if (cfa->tok[idx] == NULL) {
		ck_printf(CK_PASS,
		    "Expected unsigned integer argument\n");
		return (-1);
	}
	ep = NULL;
	ul = strtoul(cfa->tok[idx], &ep, 0);
	if (ep != NULL && *ep != '\0') {
		ck_printf(CK_PASS,
		    "Argument (\"%s\") is not an unsigned integer\n",
		    cfa->tok[idx]);
		return (-1);
	}
	**dest = ul;
	return (0);
}

int
__ck_real(CK_ARG, double **dest, int idx)
{
	double f;
	char *ep;

	if (nulled(CK_PASS, idx)) {
		*dest = NULL;
		return (0);
	}
	if (cfa->tok[idx] == NULL) {
		ck_printf(CK_PASS,
		    "Expected floating point number argument\n");
		return (-1);
	}
	ep = NULL;
	f = strtod(cfa->tok[idx], &ep);
	if (ep != NULL && *ep != '\0') {
		ck_printf(CK_PASS,
		    "Argument (\"%s\") is not a floating point numer\n",
		    cfa->tok[idx]);
		return (-1);
	}
	**dest = f;
	return (0);
}

int
__ck_follow(CK_ARG, struct __ck_node **nodes)
{
	void *priv;
	char *tree;

	tree = cfa->tree;
	if (nodes == NULL) {
		ck_printf(CK_PASS, "No configuration configured\n");
		return (-1);
	}
	if (cfa->tok[0] == NULL && (cfa->flag & (CK_CONF|CK_SHOW))) {
		priv = cfa->priv;
		for (;*nodes != NULL; nodes++) {
			/* cfa->priv must be invariant for all nodes */
			cfa->priv = priv;
			asprintf(&cfa->tree, "%s %s", tree, (*nodes)->word);
			if ((*nodes)->func != NULL)
				(*nodes)->func(cfa);
			free(cfa->tree);
		}
		cfa->tree = tree;
		return (0);
	}
	if (cfa->tok[0] == NULL && (cfa->flag & CK_HELP)) {
		for (;*nodes != NULL; nodes++) {
			ck_printf(CK_PASS,
			    "%s\t%s\n", (*nodes)->word, (*nodes)->desc);
		}
		return (0);
	}
	if (cfa->tok[0] == NULL) {
		ck_printf(CK_PASS, "Command incomplete\n");
		return (0);
	}
	for (;*nodes != NULL; nodes++) {
		if (strcmp((*nodes)->word, cfa->tok[0]))
			continue;
		cfa->tok++;
		asprintf(&cfa->tree, "%s %s", tree, (*nodes)->word);
		free(tree);
		if (cfa->tok[0] == NULL && (cfa->flag & CK_SHOW) &&
		    (*nodes)->show != NULL)
			return ((*nodes)->show(cfa));
		else if ((*nodes)->func != NULL)
			return ((*nodes)->func(cfa));
	}
	ck_printf(CK_PASS, "Syntax error (%s)\n", cfa->line);
	return (-1);
}

int
__ck_ellipsis(CK_ARG, char ***dest, int idx)
{
	int i;

	*dest = NULL;
	for (i = 0; i < idx; i++) {
		if (cfa->tok[i] != NULL)
			continue;
		if ((cfa->flag & (CK_NO|CK_CONF|CK_SHOW)))
			return (0);
		else
			return (1);
	}
	*dest = cfa->tok + idx;
	return (0);
}

/*
 * Function for recursing into another tree
 */

int
ck_continue(CK_ARG, struct __ck_node *n, char **argv)
{

	if (argv != NULL)
		cfa->tok = argv;
	else
		cfa->tok[0] = NULL;

	if ((cfa->flag & CK_SHOW) && n->show != NULL)
		return (n->show(cfa));
	if (n->func != NULL)
		return (n->func(cfa));

	ck_printf(CK_PASS, "Huh (%s@%s:%d)?\n",
	    __func__, __FILE__, __LINE__);
	return (0);
}
