/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: main.c,v 1.37 2008/02/16 12:53:00 phk Exp $
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <stdarg.h>
#include "defs.h"

extern FILE *yyin, *yyout;

static FILE *fc, *fh;

int lineno;

static int i_flag;

int
yywrap(void) {
	return (1);
}

void
yyerror(const char *fmt, ...)
{
	va_list ap;
	char msg[128];

	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);
	errx(2, "%s around line %d", msg, lineno);
}

struct info *
new_info(const char *label, const char *contents)
{
	struct info *ip;

	ip = calloc(sizeof *ip, 1);
	ip->label = strdup(label);
	if (contents != NULL)
		ip->contents = strdup(contents);
	return (ip);
}

struct infolist *
new_infolist(void)
{
	struct infolist *il;

	il = calloc(sizeof *il, 1);
	TAILQ_INIT(il);
	return (il);
}

void
append_infolist(struct infolist *il, const char *label, const char *contents)
{
	struct info *ip;

	ip = new_info(label, contents);
	TAILQ_INSERT_TAIL(il, ip, list);
}


static const char *
findinfo(struct infolist *list, const char *label)
{
	struct info *ip;

	TAILQ_FOREACH(ip, list, list)
		if (!strcmp(ip->label, label))
			return (ip->contents);
	return (NULL);
}

static const char *
c_type_of(const char *type)
{
	if (!strcmp(type, "UINT"))
		return ("unsigned");
	if (!strcmp(type, "INT"))
		return ("int");
	if (!strcmp(type, "REAL"))
		return ("double");
	if (!strcmp(type, "ELLIPSIS"))
		return ("char *");
	return ("const char *");
}

static const char *
c_func_of(const char *type)
{
	if (!strcmp(type, "UINT"))
		return ("__ck_uint");
	if (!strcmp(type, "INT"))
		return ("configkit_int");
	if (!strcmp(type, "REAL"))
		return ("__ck_real");
	if (!strcmp(type, "WORD"))
		return ("__ck_word");
	if (!strcmp(type, "ELLIPSIS"))
		return ("__ck_ellipsis");
	if (!strcmp(type, "STRING"))
		return ("configkit_string");
	return (type);
}

const char *
emit_node(const char *id, struct infolist *list)
{
	struct info *ip;
	char *name;
	const char *func;
	const char *desc;
	const char *hide;
	const char *show;

	show = findinfo(list, "show");
	if (show == NULL)
		show = "NULL";
	else
		fprintf(fh, "int %s(CK_ARG);\n", show);
	func = findinfo(list, "name");
	if (func == NULL) {
		asprintf(&name, "__node_%s_at_line_%d", id, lineno);
		hide = "static ";
	} else {
		asprintf(&name, "%s", func);
		if (!i_flag) {
			fprintf(fh, "extern struct __ck_node %s;\n", name); 
			hide = "";
		} else {
			hide = "static";
		}
	}
	fprintf(fc, "// emit_node(%d, %s)\n", lineno, id);

	fprintf(fc, "static struct __ck_node *%s_entries[] = {\n", name);
	TAILQ_FOREACH(ip, list, list) {
		if (strcmp(ip->label, "ENTRY"))
			continue;
		fprintf(fc, "\t&%s,\n", ip->contents);
	}
	fprintf(fc, "\tNULL\n");
	fprintf(fc, "\t};\n\n");

	fprintf(fc, "static int\n");
	fprintf(fc, "%s_func(CK_ARG)\n", name);
	fprintf(fc, "{\n");
	fprintf(fc, "\n");
	func = findinfo(list, "func");
	if (func != NULL)
		errx(1, "extra func in %s around line %d\n", id, lineno);

	fprintf(fc,
	    "\treturn(__ck_follow(CK_PASS, %s_entries));\n",
	    name);

	fprintf(fc, "}\n\n");

	fprintf(fc, "%sstruct __ck_node %s = {\n", hide, name);
	fprintf(fc, "\t\"%s\",\n", id);
	fprintf(fc, "\t%s_func,\n", name);
	fprintf(fc, "\t%s,\n", show);
	desc = findinfo(list, "desc");
	if (desc == NULL)
		errx(1, "Missing description in \"%s\" around line %d\n",
		   id, lineno);
	fprintf(fc, "\t\"%s\"\n", desc);
	fprintf(fc, "};\n\n");

	return (name);
}


const char *
emit_word(const char *id, struct infolist *type_list, struct infolist *list)
{
	struct info *ip;
	char *name;
	const char *show;
	const char *func;
	const char *desc;
	const char *hide;
	int narg;
	int had_ellipsis;

	show = findinfo(list, "show");
	if (show == NULL)
		show = "NULL";
	else
		fprintf(fh, "int %s(CK_ARG);\n", show);
	func = findinfo(list, "name");
	if (func == NULL) {
		asprintf(&name, "__word_%s_at_line_%d", id, lineno);
		hide = "static ";
	} else {
		name = strdup(func);
		hide = "";
		fprintf(fh, "extern struct __ck_node %s;\n", name);
	}
	fprintf(fc, "// emit_word(%d, %s)\n", lineno, id);
	fprintf(fc, "static int\n");
	fprintf(fc, "%s_func(CK_ARG)\n", name);
	fprintf(fc, "{\n");
	fprintf(fc, "\n");
	narg = 0;
	had_ellipsis = 0;
	TAILQ_FOREACH(ip, type_list, list) {
		fprintf(fc, "\t{\n");
		fprintf(fc, "\t%s arg_%d;\n", c_type_of(ip->label), narg);
		fprintf(fc, "\t%s *parg_%d = &arg_%d;\n",
		    c_type_of(ip->label), narg, narg);
		fprintf(fc, "\tif (%s(CK_PASS,", c_func_of(ip->label));
		fprintf(fc, " &parg_%d, %d))\n", narg, narg);
		fprintf(fc, "\t\treturn (-1);\n");
		narg++;
		if (!strcmp(ip->label, "ELLIPSIS"))
			had_ellipsis = 1;
	}

	if (!had_ellipsis) {
		fprintf(fc, "\tif (!(cfa->flag & (CK_NO|CK_SHOW|CK_CONF)) ");
		fprintf(fc, "&& cfa->tok[%d] != NULL) {\n", narg);
		fprintf(fc, "\t\tck_printf(CK_PASS, \"%s\\n\");",
		    "Too many arguments");
		fprintf(fc, "\t\treturn (-1);\n");
		fprintf(fc, "\t}\n");
	}
	
	func = findinfo(list, "func");
	if (narg > 0 && func == NULL)
		errx(1, "missing func in %s around line %d\n", id, lineno);
	if (func != NULL) {
		fprintf(fh, "int %s(CK_ARG", func);
		fprintf(fc, "\treturn (%s(CK_PASS", func);
		narg = 0;
		TAILQ_FOREACH(ip, type_list, list) {
			fprintf(fc, ", parg_%d", narg);
			fprintf(fh, ", %s *arg_%d", c_type_of(ip->label), narg);
			narg++;
		}
		fprintf(fc, "));\n");
		fprintf(fh, ");\n");
	}
	TAILQ_FOREACH(ip, type_list, list) {
		fprintf(fc, "\t}\n");
	}

	fprintf(fc, "}\n\n");

	fprintf(fc, "%sstruct __ck_node %s = {\n", hide, name);
	fprintf(fc, "\t\"%s\",\n", id);
	fprintf(fc, "\t%s_func,\n", name);
	fprintf(fc, "\t%s,\n", show);
	desc = findinfo(list, "desc");
	if (desc == NULL)
		errx(1, "Missing description in \"%s\" around line %d\n",
		   id, lineno);
	fprintf(fc, "\t\"%s\"\n", desc);
	fprintf(fc, "};\n\n");

	return (name);
}

const char *
emit_extern(const char *id, const char *ref, struct infolist *list)
{
	char *name;
	const char *setup;
	const char *desc;
	const char *hide;
	const char *func;
	const char *show;

	show = findinfo(list, "show");
	if (show == NULL)
		show = "NULL";
	else
		fprintf(fh, "int %s(CK_ARG);\n", show);
	func = findinfo(list, "name");
	if (func == NULL) {
		asprintf(&name, "__extern_%s_at_line_%d", id, lineno);
		hide = "static ";
	} else {
		name = strdup(func);
		hide = "";
		fprintf(fh, "extern struct __ck_node **%s;\n", name);
	}
	fprintf(fc, "// emit_extern(%d, %s, %s)\n", lineno, id, ref);
	fprintf(fh, "extern struct __ck_node %s;\n", ref);
	fprintf(fc, "static int\n");
	fprintf(fc, "%s_func(CK_ARG)\n", name);
	fprintf(fc, "{\n");
	fprintf(fc, "\n");
	
	setup = findinfo(list, "setup");
	if (setup != NULL) {
		fprintf(fc, "\tif (%s(CK_PASS))\n", setup);
		fprintf(fc, "\t\treturn (0);\n");
		fprintf(fh, "int %s(CK_ARG);\n", setup);
	}

	fprintf(fc,
	    "\tif ((cfa->flag & CK_SHOW) && %s.show != NULL && cfa->tok[0] == NULL)\n", ref);
	fprintf(fc,
	    "\t\treturn(%s.show(CK_PASS));\n", ref);
	
	fprintf(fc,
	    "\treturn(%s.func(CK_PASS));\n", ref);
	fprintf(fc, "}\n\n");

	fprintf(fc, "%sstruct __ck_node %s = {\n", hide, name);
	fprintf(fc, "\t\"%s\",\n", id);
	fprintf(fc, "\t%s_func,\n", name);
	fprintf(fc, "\t%s,\n", show);
	desc = findinfo(list, "desc");
	if (desc == NULL)
		errx(1, "Missing description in \"%s\" around line %d\n",
		   id, lineno);
	fprintf(fc, "\t\"%s\"\n", desc);
	fprintf(fc, "};\n\n");

	return (name);
}

const char *
emit_import(const char *id __unused, struct infolist *list __unused)
{
	return (NULL);
}

const char *
emit_instance(const char *id, const char *type, struct infolist *list)
{
	struct info *ip;
	const char *func, *ctor, *dtor, *show;
	const char *hide, *desc, *head;
	char *name, *hname;

	show = findinfo(list, "show");
	if (show != NULL)
		fprintf(fh, "int %s(CK_ARG);\n", show);
	func = findinfo(list, "name");
	ctor = findinfo(list, "new");
	head = findinfo(list, "head");
	if (ctor == NULL)
		errx(1, "Missing \"new\" in \"%s\" around line %d",
		    id, lineno);
	dtor = findinfo(list, "destroy");
	if (func == NULL) {
		asprintf(&name, "instance_%s_at_line_%d", id, lineno);
		hide = "static ";
	} else {
		name = strdup(func);
		hide = "";
		fprintf(fh, "extern struct __ck_node %s;\n", name);
	}

	if (head == NULL) {
		asprintf(&hname, "instance_head_%s_at_line_%d", id, lineno);
		fprintf(fc, "static struct ck_instancehead %s =\n",
		    hname);
		fprintf(fc, "    TAILQ_HEAD_INITIALIZER(%s);\n", hname);
	} else {
		hname = NULL;
	}

	fprintf(fc, "//emit_instance(%d, %s, %s)\n", lineno, id, type);
	TAILQ_FOREACH(ip, list, list)
		fprintf(fc, "\t// <%s> <%s>\n",
		    ip->label, ip->contents);
	fprintf(fc, "\n");

	fprintf(fc, "static struct __ck_node *%s_entries[] = {\n", name);
	TAILQ_FOREACH(ip, list, list) {
		if (strcmp(ip->label, "ENTRY"))
			continue;
		fprintf(fc, "\t&%s,\n", ip->contents);
	}
	fprintf(fc, "\tNULL\n");
	fprintf(fc, "\t};\n\n");

	fprintf(fh, "struct ck_instance *%s(CK_ARG, %s);\n", ctor, c_type_of(type));

	if (dtor != NULL)
		fprintf(fh, "int %s(CK_ARG);\n", dtor);

	if (head != NULL)
		fprintf(fh,
		    "struct ck_instancehead *%s(CK_ARG);\n", head);

	fprintf(fc, "static int\n");
	fprintf(fc, "%s_func(CK_ARG)\n", name);
	fprintf(fc, "{\n");
	fprintf(fc, "\t%s idx;\n", c_type_of(type));
	fprintf(fc, "\tstruct ck_instance *cip;\n");
	fprintf(fc, "\tstruct ck_instancehead *cihp;\n");
	fprintf(fc, "\t%s *pidx = &idx;\n", c_type_of(type));
	fprintf(fc, "\tchar *p;\n");
	fprintf(fc, "\n");
	if (head != NULL)
		fprintf(fc, "\tcihp = %s(CK_PASS);\n", head);
	else
		fprintf(fc, "\tcihp = &%s;\n", hname);
	if (show != NULL) {
		fprintf(fc, "\tif ((cfa->flag & CK_SHOW) && cfa->tok[0] == NULL) {\n");
		fprintf(fc, "\t\tp = cfa->tree;\n");
		fprintf(fc, "\t\tTAILQ_FOREACH(cip, cihp, next) {\n");
		fprintf(fc, "\t\t\tcfa->priv = cip->obj;\n");
		if (!strcmp(type, "WORD")) 
			fprintf(fc, "\t\t\tasprintf(&cfa->tree, \"%%s %%s\", p, cip->idx.word);\n");
		else
			fprintf(fc, "\t\t\tasprintf(&cfa->tree, \"%%s %%u\", p, cip->idx.uint);\n");
		fprintf(fc, "\t\t\t%s(CK_PASS);\n", show);
		fprintf(fc, "\t\t\tfree(cfa->tree);\n");
		fprintf(fc, "\t\t}\n");
		fprintf(fc, "\t\tcfa->tree = p;\n");
		fprintf(fc, "\t\treturn (0);\n");
		fprintf(fc, "\t\t}\n");
	}
	fprintf(fc, "\tif ((cfa->flag & (CK_HELP|CK_CONF|CK_SHOW)) && cfa->tok[0] == NULL)\n");
	if (!strcmp(type, "WORD")) {
		fprintf(fc, "\t\treturn (__ck_walk_word(CK_PASS,\n");
	} else if (!strcmp(type, "UINT")) {
		fprintf(fc, "\t\treturn (__ck_walk_uint(CK_PASS,\n");
	}
	fprintf(fc, "\t\t    cihp, %s_entries));\n", name);
	fprintf(fc, "\n");

	fprintf(fc, "\tif (%s(CK_PASS, &pidx, 0))\n", c_func_of(type));
	fprintf(fc, "\t\treturn (-1);\n");
	fprintf(fc, "\tcfa->tok++;\n");
	fprintf(fc, "\n");

	fprintf(fc, "\tTAILQ_FOREACH(cip, cihp, next) {\n");
	fprintf(fc, "\t\tif (cip->parent != cfa->priv)\n");
	fprintf(fc, "\t\t\tcontinue;\n");
	if (!strcmp(type, "WORD")) {
		fprintf(fc, "\t\tif (!strcmp(cip->idx.word, idx)) {\n");
	} else if (!strcmp(type, "UINT")) {
		fprintf(fc, "\t\tif (cip->idx.uint == idx) {\n");
	}
	fprintf(fc, "\t\t\tp = cfa->tree;\n");
	if (!strcmp(type, "WORD")) {
		fprintf(fc, "\t\t\tasprintf(&cfa->tree, \"%%s %%s\", p, cip->idx.word);\n");
	} else if (!strcmp(type, "UINT")) {
		fprintf(fc, "\t\t\tasprintf(&cfa->tree, \"%%s %%u\", p, cip->idx.uint);\n");
	}
	fprintf(fc, "\t\t\tfree(p);\n");
	fprintf(fc, "\t\t\tbreak;\n");
	fprintf(fc, "\t\t}\n");
	fprintf(fc, "\t}\n");

	fprintf(fc, "\tif (cip == NULL && (cfa->flag & (CK_NO|CK_CONF|CK_SHOW)))\n");
	fprintf(fc, "\t\treturn (0);\n");

	if (dtor != NULL) {
		fprintf(fc, "\tif (cip != NULL && cfa->tok[0] == NULL && (cfa->flag & CK_NO)) {\n");
		fprintf(fc, "\t\tcfa->priv = cip->obj;\n");
		fprintf(fc, "\t\tTAILQ_REMOVE(cihp, cip, next);\n");
		fprintf(fc, "\t\treturn (%s(CK_PASS));\n", dtor);
		fprintf(fc, "\t}\n");
	}

	fprintf(fc, "\tif (cip == NULL) {\n");
	fprintf(fc, "\t\tcip = %s(CK_PASS, idx);\n", ctor);
	fprintf(fc, "\t\tif (cip == NULL)\n");
	fprintf(fc, "\t\t\treturn (0);\n");
	if (!strcmp(type, "WORD")) {
		fprintf(fc, "\t\tcip->idx.word = strdup(idx);\n");
		
	} else if (!strcmp(type, "UINT")) {
		fprintf(fc, "\t\tcip->idx.uint = idx;\n");
	}
	fprintf(fc, "\t\tcip->parent = cfa->priv;\n");
	fprintf(fc, "\t\tTAILQ_INSERT_TAIL(cihp, cip, next);\n");
	fprintf(fc, "\t}\n");
	fprintf(fc, "\tcfa->priv = cip->obj;\n");
	fprintf(fc, "\n");

	if (show != NULL) {
		fprintf(fc, "\tif (cfa->tok[0] == NULL && cfa->flag & CK_SHOW)\n");
		fprintf(fc, "\t\treturn (%s(CK_PASS));\n", show);
		fprintf(fc, "\n");
	}

	fprintf(fc, "\tif (cfa->tok[0] == NULL &&\n");
	fprintf(fc, "\t    !(cfa->flag & (CK_HELP|CK_CONF)))\n");
	fprintf(fc, "\t\treturn (0);\n");


	fprintf(fc,
	    "\treturn(__ck_follow(CK_PASS,\n");
	fprintf(fc,
	    "\t    %s_entries));\n", name);

	fprintf(fc, "}\n\n");

	fprintf(fc, "%sstruct __ck_node %s = {\n", hide, name);
	fprintf(fc, "\t\"%s\",\n", id);
	fprintf(fc, "\t%s_func,\n", name);
	fprintf(fc, "\tNULL,\n");
	desc = findinfo(list, "desc");
	if (desc == NULL)
		errx(1, "Missing description in \"%s\" around line %d\n",
		   id, lineno);
	fprintf(fc, "\t\"%s\"\n", desc);
	fprintf(fc, "};\n\n");


	free(hname);
	return (name);
}

const char *
emit_enum(struct infolist *list)
{
	struct info *ip;
	char *name;

	asprintf(&name, "__enum_at_line_%d", lineno);
	fprintf(fc, "// emit_enum(%d, %p)\n", lineno, list);
	fprintf(fc, "static const char *%s_table[] = {\n", name);
	TAILQ_FOREACH(ip, list, list)
		fprintf(fc, "\t\"%s\",\n", ip->label);
	fprintf(fc, "\tNULL\n");
	fprintf(fc, "};\n\n");

	fprintf(fc, "static int\n");
	fprintf(fc, "%s(CK_ARG, const char ***dest, int idx)\n", name);
	fprintf(fc, "{\n");
	fprintf(fc, "\n");
	fprintf(fc,
	    "\treturn(__ck_enum(CK_PASS, dest, idx, %s_table));\n",
	    name);
	fprintf(fc, "}\n\n");

	return (name);
}

int
main(int argc, char **argv)
{
	char *p, *c, *h;
	int ch;

	while ((ch = getopt(argc, argv, "i")) != -1) {
		switch (ch) {
		case 'i':
			i_flag = 1;
			break;
		default:
			errx(1, "Usage!");
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 1)
		errx(1, "Usage!");
	if (argc != 0) {
		yyin = fopen(argv[0], "r" );
		p = strdup(argv[0]);
	} else {
		yyin = stdin;
		p = strdup("configkit_conf.ck");
	}

	c = strchr(p, '\0');
	if (c >= p + 3 && c[-3] == '.' && c[-2] == 'c' && c[-1] == 'k') {
		c[-3] = '\0';
	}
	c = strrchr(p, '/');
	if (c != NULL)
		p = c + 1;

	if (i_flag) {
		asprintf(&c, "/tmp/%d.c", getpid());
		asprintf(&h, "/tmp/%d.h", getpid());
	} else {
		asprintf(&c, "%s.c", p);
		asprintf(&h, "%s.h", p);
	}

	fc = fopen(c, "w");
	setbuf(fc, NULL);
	if (!i_flag) {
		fprintf(fc, "#include <stdio.h>\n");
		fprintf(fc, "#include <stdlib.h>\n");
		fprintf(fc, "#include <string.h>\n");
		/* fprintf(fc, "#include <sys/types.h>\n"); */
	}
	fprintf(fc, "#include <configkit.h>\n");
	fprintf(fc, "#include <configkit_int.h>\n");
	if (!i_flag) {
		fprintf(fc, "#include \"%s\"\n", h);
	}

	fh = fopen(h, "w");
	setbuf(fh, NULL);

	yyparse();
	fclose(fc);
	fclose(fh);
	if (i_flag) {
		asprintf(&p, "cat %s %s", h, c);
		assert(p != NULL);
		system(p);
		unlink(c);
		unlink(h);
	}
	
	return (0);
}
